import datetime
import json
import logging
import random
import threading
import traceback

import pika

from clc_msa_utils.kv_store import KVStore
from clc_msa_utils.log_manager import LogManager


class QueueConsumer:

    def __init__(self, config={}):
        self._config = {}
        self._logger = logging.getLogger('QueueConsumer')
        self._connection = None
        self._channel = None
        self._callback = None
        self._worker_thread = None
        self._stop_consuming = False

        self._config = config

    def _configure_listener(self):
        if 'auth' not in self._config \
                or 'user' not in self._config['auth'] \
                or 'password' not in self._config['auth'] \
                or 'queue' not in self._config \
                or 'host' not in self._config \
                or 'port' not in self._config:
            self._logger.error("config=" + str(self._config))
            raise Exception("Expected the consumer configuration to have an auth.user, "
                            "auth.password, host, queue, and port attribute, but received: {0}"
                            .format(str(self._config)))
        credentials = pika.PlainCredentials(self._config['auth']['user'], self._config['auth']['password'])
        parameters = pika.ConnectionParameters(self._config['host'], int(self._config['port']), '/', credentials)
        self._logger.debug("Creating connection.")
        self._connection = pika.BlockingConnection(parameters)
        if not self._stop_consuming:
            self._logger.debug("Creating channel.")
            self._channel = self._connection.channel()
            self._logger.debug("Declaring queue.")
            self._channel.queue_declare(queue=self._config['queue'], durable=True)

            if 'exchange' in self._config:
                self._channel.queue_bind(queue=self._config['queue'], exchange=self._config['exchange']['name'],
                                         routing_key=self._config['binding_key'])

            self._logger.debug("Set listener on queue {0}."
                               .format(self._config['queue']))
            self._channel.basic_qos(prefetch_count=1)
            self._channel.basic_consume(self._config['queue'], self._callback)

            self._logger.debug("Start consuming on queue {0}."
                               .format(self._config['queue']))
            self._channel.start_consuming()

    def get_config(self):
        return self._config

    def listen(self, callback, blocking=True):
        self._callback = callback

        if blocking:
            self._logger.debug("Creating blocking listener.")
            self._configure_listener()
        else:
            self._logger.debug("Creating non-blocking listener.")
            self._worker_thread = threading.Thread(target=self._configure_listener)
            self._worker_thread.start()

    def stop_consuming(self):
        self._stop_consuming = True
        try:
            if not self._channel:
                self._logger.debug("Channel has not start consuming yet.".format(self._config.get('name')))
                return
            if not self._channel.is_closed:
                try:
                    self._channel.stop_consuming()
                    self._logger.debug("Channel stopped consuming on '{0}' thread.".format(self._config.get('name')))
                except:
                    self._logger.warning(traceback.format_exc())
                    self._logger.warning("Channel wouldn't stop consuming on '{0}' thread, closing channel.".format(
                        self._config.get('name')))
                    self._channel.close()
        except:
            self._logger.error(traceback.format_exc())

    def thread(self):
        return self._worker_thread

    def config(self):
        return self._config

    def callback(self):
        return self._callback


class QueueProducer:

    def __init__(self, config={}):
        self._config = {}
        self._logger = logging.getLogger('QueueProducer')
        self._connection = None
        self._queue_channel = None
        self._config = self._normalize_config(config)
        self._configure_producer()

    def _configure_producer(self):
        self._logger.debug("Configuring queue producer with config:\n%s" % self._config)
        if not self._connection or self._connection.is_closed:
            credentials = pika.PlainCredentials(self._config['auth']['user'], self._config['auth']['password'])
            parameters = pika.ConnectionParameters(self._config['host'], int(self._config['port']), '/', credentials)
            self._connection = pika.BlockingConnection(parameters)
        if not self._queue_channel or self._queue_channel.is_closed:
            self._queue_channel = self._connection.channel()

            if 'create' in self._config['exchange'] and self._config['exchange']['create']:
                self._logger.debug("Exchange is being ensured present based on configuration...")
                self._queue_channel \
                    .exchange_declare(exchange=self._config['exchange']['name'],
                                      exchange_type=(
                                          self._config['exchange']['type'] if 'type' in self._config[
                                              'exchange'] else 'direct'),
                                      arguments=(self._config['exchange']['arguments'] if 'arguments' in
                                                                                          self._config[
                                                                                              'exchange'] else {}),
                                      durable=(
                                          self._config['exchange']['durable'] if 'durable' in self._config[
                                              'exchange'] else True))

    def _normalize_config(self, config):
        return config

    def publish(self, message=None, properties=None, try_number=0):
        try_number = try_number + 1

        self._logger.info("Publishing message to exchange '{0}' with binding key '{1}'"
                          .format(self._config['exchange']['name'],
                                  self._config['binding_key']))
        self._configure_producer()

        if type(message) is dict:
            body = json.dumps(message)
        else:
            body = str(message)

        self._logger.debug("Message:\n%s" % body)

        if not properties:
            properties = pika.BasicProperties()

        # Make message persistent
        properties.delivery_mode = 2
        # Assume json
        properties.content_type = 'application/json'

        if self._connection.is_closed or self._queue_channel.is_closed:
            self._configure_producer()

        try:
            self._queue_channel.basic_publish(exchange=self._config['exchange']['name'],
                                              routing_key=(
                                                  self._config[
                                                      'binding_key'] if 'binding_key' in self._config else '#'),
                                              body=body,
                                              properties=properties)
            self._logger.info("SUCCESS: Message Published")
        except Exception as e:
            if try_number > 5:
                raise e
            else:
                self._logger.info("FAILED: An error occurred publishing to the {0} exchange with binding key {1},"
                                  " retrying. This is try number {2}."
                                  .format(self._config['exchange']['name'],
                                          self._config['binding_key'] if 'binding_key' in self._config else '#',
                                          str(try_number)))
                self._configure_producer()
                self.publish(message=message, properties=properties, try_number=try_number)

    def get_config(self):
        return self._config


class QueueFactory:
    """
    QueueFactory provides helper methods for simply passing in a configuration
    and having producers/consumers created based on that config.  Additionally,
    there will be builder style methods available for setting/overriding configuration
    prior to generating producers/consumers.

    Currently, this solution is satisfying making a connection using AMQP which is compatible with
    RabbitMQ and ActiveMQ.  It could be enhanced to provide further abstraction and provide
    connections to other Queue services in theory.
    """

    _logger = logging.getLogger('QueueFactory')

    def __init__(self, monitoring_period_seconds=10):
        self._consumers = {}
        self._producers = {}
        self._config = {}
        self._monitoring = False
        self._monitoring_period_seconds = monitoring_period_seconds
        self._monitoring_thread = None
        self._logger.info("Initializing QueueFactory...")

    def create_consumers(self, queue_config):
        """
        Example Consumer Queue Configuration:
        {
          "queue_config": [
            {
              "name": "name-to-reference-this-connection",
              "type": "consumer",
              "exchange": {
                "name": "name-of-exchange",
                "type": "direct | fanout | etc",
                "arguments": {"x-delayed-type": "topic"},
                "durable": true | false
              },
              "queue": "name-of-queue",
              "binding_key": "#",
              "host": "rabbitmq-host",
              "port": "5672",
              "auth": {
                "user": "guest",
                "password": "guest"
              }
            }
          ]
        }
        """

        if 'queue_config' in queue_config:
            queue_config = queue_config['queue_config']

        self._logger.debug("Setting up consumers for the following configuration:\n%s" % str(queue_config))
        for config in queue_config:
            if 'type' in config and 'consumer' == config['type']:
                consumer = QueueConsumer(config)
                self._logger.debug("Adding configuration for configuration %s." % str(config['name']))
                self._consumers[config['name']] = consumer

        self._logger.debug("Created {0} consumers.".format(str(len(self._consumers))))
        for consumer in list(self._consumers.values()):
            self._logger.debug("Created {0} consumer.".format(consumer.config().get("name")))
        return self._consumers

    def create_producers(self, queue_config):
        """
        Example Producer Queue Configuration:
        {
            "queue_config": [
                {
                    "name": "name-to-reference-this-connection",
                    "type": "producer",
                    "exchange": {
                        "name": "name-of-exchange",
                        "create": true | false (whether or not to ensure exchange is present),
                        "type": "direct | fanout | etc",
                        "arguments": {"x-delayed-type": "topic"},
                        "durable": true | false
                    },
                    "binding_key": "#",
                    "host": "rabbitmq-host",
                    "port": "5672",
                    "auth": {
                        "user": "guest",
                        "password": "guest"
                    }
                }
            ]
        }
        """

        self._logger.debug("Setting up producers for the following configuration:\n%s" % queue_config)
        if 'queue_config' in queue_config:
            queue_config = queue_config['queue_config']

        for config in queue_config:
            if 'type' in config['exchange'] and 'producer' == config['exchange']['type']:
                producer = QueueProducer(config)
                self._producers[config['name']] = producer

        return self._producers

    def stop_consuming(self):
        """
        Stops all QueueConsumers from consuming.
        """
        self.stop_monitoring()

        for consumer in list(self._consumers.values()):
            consumer.stop_consuming()

        for consumer in list(self._consumers.values()):
            consumer.stop_consuming()
        self._consumers = {}

    def start_consuming(self, callback):
        """
        Starts all QueueConsumers consuming in a separate thread.
        """
        for consumer in list(self._consumers.values()):
            consumer.listen(callback, blocking=False)
            consumer.thread().name = "WorkerThread-{0}-{1}" \
                .format(datetime.datetime.now().strftime("%d-%m-%Y %H:%M:%S"),
                        self._random_string())
        self.start_monitoring()

    def consumers(self):
        """
        Gets a list of all QueueConsumers

        :return: list of QueueConsumers
        """
        return list(self._consumers.values())

    def start_monitoring(self):
        self._logger.debug("Thread Monitoring Enabled checking threads every {0} seconds."
                           .format(self._monitoring_period_seconds))
        self._monitoring = True
        self._monitoring_thread = threading.Timer(self._monitoring_period_seconds, self._monitor)
        self._logger.debug("Created monitoring thread named {0}."
                           .format(self._monitoring_thread.name))
        self._monitoring_thread.start()
        self._monitoring_thread.name = "MonitorThread-{0}-{1}" \
            .format(datetime.datetime.now().strftime("%d-%m-%Y %H:%M:%S"),
                    self._random_string())

        self._logger.debug("Monitoring thread started and is alive? {0}."
                           .format(str(self._monitoring_thread.is_alive())))

    def _random_string(self):
        return str(random.randrange(0, 999999)).zfill(6)

    def monitoring_thread(self):
        return self._monitoring_thread

    def stop_monitoring(self):
        self._logger.debug("Thread Monitoring Disabled")
        self._monitoring = False

    def _monitor(self):
        if self._logger.isEnabledFor(logging.DEBUG):
            self._logger.debug("Monitoring {0} threads..."
                               .format(str(len(list(self._consumers.values())))))
            self._logger.debug("There are currently {0} active threads in the system."
                               .format(str(threading.active_count())))
            for thread in threading.enumerate():
                self._logger.debug("Thread [{0}] is alive.".format(thread.name))
        if self._monitoring:
            for consumer in list(self._consumers.values()):
                if not self._monitoring:
                    break
                if not consumer.thread().is_alive():
                    consumer.listen(consumer.callback(), blocking=False)
                    self._logger.debug("Consumer thread {0} was not alive, successfully started"
                                       .format(consumer.config().get("name")))
                else:
                    self._logger.debug("Consumer thread {0} is still active."
                                       .format(consumer.config().get("name")))

        if self._monitoring:
            self.start_monitoring()
        self._logger.debug("Done monitoring threads.")


class QueueWorker:
    """
        Helper class to make writing queue workers easy. Specify kv store connection info,
        keys to retrieve configuration from the kv_store for the queue connection and initialization,
        and write your listener method. This class creates a KVStore and LogManager for you.
    """

    _logger = logging.getLogger("QueueWorker")

    def __init__(self,
                 consul_host=None,
                 consul_port=None,
                 etcd_host=None,
                 etcd_port=None,
                 kv_prefix=None,
                 rabbit_host_key="rabbit_host",
                 rabbit_port_key="rabbit_port",
                 rabbit_user_key="rabbit_user",
                 rabbit_password_key="rabbit_password",
                 amqp_connection_key="amqp_connections",
                 queue_name_key="queue_name",
                 listen_exchange_key="exchange",
                 listen_routing_key_key="listen_routing_key",
                 done_exchange_key="done_exchange",
                 done_routing_key_key="done_routing_key",
                 error_exchange_key="error_exchange",
                 error_routing_key_key="error_routing_key",
                 initialize_log_manager=True,
                 kv_store=None,
                 data_key_on_error_payload="data",

                 rabbit_host_default="localhost",
                 rabbit_port_default="5672",
                 rabbit_user_default="guest",
                 rabbit_password_default="guest",
                 queue_name_default="default_queue",
                 amqp_connection_default="10",
                 listen_exchange_default="main_exchange",
                 listen_routing_default="listen.key",
                 done_exchange_default="main_exchange",
                 done_routing_key_default="done.key",
                 error_exchange_default="error_exchange",
                 error_routing_key_default="error.key"):

        # Initialize instance variables.
        self._consul_host = consul_host
        self._consul_port = consul_port
        self._etcd_host = etcd_host
        self._etcd_port = etcd_port
        self._kv_prefix = kv_prefix

        # RabbitMQ connection, exchange, queue, and binding key KVStore keys
        self._rabbit_host_key = rabbit_host_key
        self._rabbit_port_key = rabbit_port_key
        self._rabbit_user_key = rabbit_user_key
        self._rabbit_password_key = rabbit_password_key
        self._amqp_connection_key = amqp_connection_key
        self._listen_exchange_key = listen_exchange_key
        self._queue_name_key = queue_name_key
        self._listen_routing_key_key = listen_routing_key_key
        self._done_exchange_key = done_exchange_key
        self._done_routing_key_key = done_routing_key_key
        self._error_exchange_key = error_exchange_key
        self._error_routing_key_key = error_routing_key_key
        self._error_original_payload_key = data_key_on_error_payload

        # RabbitMQ default values
        self._rabbit_host_default = rabbit_host_default
        self._rabbit_port_default = rabbit_port_default
        self._rabbit_user_default = rabbit_user_default
        self._rabbit_password_default = rabbit_password_default
        self._queue_name_default = queue_name_default
        self._amqp_connection_default = amqp_connection_default
        self._listen_exchange_default = listen_exchange_default
        self._listen_routing_default = listen_routing_default
        self._done_exchange_default = done_exchange_default
        self._done_routing_key_default = done_routing_key_default
        self._error_exchange_default = error_exchange_default
        self._error_routing_key_default = error_routing_key_default

        # Other private variables
        self._callback = None
        self._done_producer = None
        self._error_producer = None
        self._queue_factory = None

        if kv_store:
            self._kv_store = kv_store
        else:
            self._kv_store = KVStore(
                consul_host=consul_host,
                consul_port=consul_port,
                etcd_host=etcd_host,
                etcd_port=etcd_port,
                kv_prefix=kv_prefix,
                reload_enabled=True
            )

        if initialize_log_manager:
            self._log_manager = LogManager(kv_store=self._kv_store)

        self._logger.info("QueueWorker initialized.")

    def kv_store(self):
        """
            Returns the KVStore created by this class.
            :return: KVStore
        """
        return self._kv_store

    def set_callback(self, callback):
        """
            Sets the callback for the queue worker, and starts consuming
            :param callback:
        """
        self._logger.info("Setting callback...")
        self._callback = callback
        self._kv_store.on_reload(self._kv_store_on_reload)
        self._logger.info("Start consuming on queue {0}..."
                          .format(self._kv_store.get(self._queue_name_key,
                                                     self._queue_name_default)))
        self._setup_publishers_and_consumers({}, {}, force=True)
        self._logger.info("Consuming messages on queue {0}."
                          .format(self._kv_store.get(self._queue_name_key,
                                                     self._queue_name_default)))

    def publish_success(self, data, properties=None):
        """
            Publishes a message to the success queue.

            :param data: The payload.
            :param properties: The properties for the payload.
        """
        self._done_producer.publish(data, properties)

    def publish_error(self, data, properties=None):
        """
            Publishes a message to the error queue.

            :param data: The payload.
            :param properties: The properties for the payload.
        """
        self._error_producer.publish(data, properties)

    def _callback_wrapper(self, ch, method, properties, body):
        """
        Calls the callback method, handles Exceptions, publishes error details to the error queue, an acknowledges
        errors.
        :param ch:
        :param method:
        :param properties:
        :param body:
        :return:
        """
        self._logger.debug("Calling callback...")
        try:
            self._callback(ch, method, properties, body)
        except Exception as exception:
            error_message = "Error while processing message on the `{0}` exchange with a `{1}` binding key." \
                .format(method.exchange, method.routing_key)

            self._logger.error(error_message)
            stacktrace = traceback.format_exc()
            self._logger.error(exception)
            self._logger.error(stacktrace)

            try:
                message = json.loads(body.decode("utf-8"))
            except:
                message = body.decode("utf-8")

            error_body = {
                self._error_original_payload_key: message,
                'message': stacktrace.splitlines()[-1] if stacktrace else error_message,
                'error_details': stacktrace,
                'errorDetails': stacktrace
            }
            self.publish_error(json.dumps(error_body), properties)

        ch.basic_ack(delivery_tag=method.delivery_tag)
        self._logger.debug("Done calling callback")

    def _kv_store_on_reload(self, old_config, new_config):
        self._setup_publishers_and_consumers(old_config, new_config)

    def _setup_publishers_and_consumers(self, old_config, new_config, force=False):

        if force or self._kv_store.attribute_changed_in_dict(old_config, new_config,
                                                             self._rabbit_host_key,
                                                             self._rabbit_port_key,
                                                             self._rabbit_user_key,
                                                             self._rabbit_password_key,
                                                             self._done_exchange_key,
                                                             self._done_routing_key_key,
                                                             self._error_exchange_key,
                                                             self._error_routing_key_key,
                                                             self._queue_name_key,
                                                             self._listen_exchange_key,
                                                             self._listen_routing_key_key,
                                                             self._amqp_connection_key):
            self._logger.info("Initializing RabbitMQ pubilishers and consumers...")
            # Setup done producer
            self._done_producer = self._new_queue_producer(self._kv_store.get(self._rabbit_host_key,
                                                                              self._rabbit_host_default),
                                                           self._kv_store.get(self._rabbit_port_key,
                                                                              self._rabbit_port_default),
                                                           self._kv_store.get(self._rabbit_user_key,
                                                                              self._rabbit_user_default),
                                                           self._kv_store.get(self._rabbit_password_key,
                                                                              self._rabbit_password_default),
                                                           self._kv_store.get(self._done_exchange_key,
                                                                              self._done_exchange_default),
                                                           self._kv_store.get(self._done_routing_key_key,
                                                                              self._done_routing_key_default))

            # Setup error producer
            self._error_producer = self._new_queue_producer(self._kv_store.get(self._rabbit_host_key,
                                                                               self._rabbit_host_default),
                                                            self._kv_store.get(self._rabbit_port_key,
                                                                               self._rabbit_port_default),
                                                            self._kv_store.get(self._rabbit_user_key,
                                                                               self._rabbit_user_default),
                                                            self._kv_store.get(self._rabbit_password_key,
                                                                               self._rabbit_password_default),
                                                            self._kv_store.get(self._error_exchange_key,
                                                                               self._error_exchange_default),
                                                            self._kv_store.get(self._error_routing_key_key,
                                                                               self._error_routing_key_default))
            # Setup queue consumers
            if self._queue_factory:
                self._queue_factory.stop_consuming()

            self._queue_factory = self._new_queue_factory(self._kv_store.get(self._rabbit_host_key,
                                                                             self._rabbit_host_default),
                                                          self._kv_store.get(self._rabbit_port_key,
                                                                             self._rabbit_port_default),
                                                          self._kv_store.get(self._rabbit_user_key,
                                                                             self._rabbit_user_default),
                                                          self._kv_store.get(self._rabbit_password_key,
                                                                             self._rabbit_password_default),
                                                          self._kv_store.get(self._listen_exchange_key,
                                                                             self._listen_exchange_default),
                                                          self._kv_store.get(self._queue_name_key,
                                                                             self._queue_name_default),
                                                          self._kv_store.get(self._listen_routing_key_key,
                                                                             self._listen_routing_default),
                                                          self._kv_store.get(self._amqp_connection_key,
                                                                             self._amqp_connection_default),
                                                          self._callback)
            self._logger.info("Done Initializing RabbitMQ pubilishers and consumers.")

    def _new_queue_factory(self,
                           rabbit_host,
                           rabbit_port,
                           rabbit_user,
                           rabbit_password,
                           exchange,
                           queue_name,
                           binding_key,
                           amqp_connections,
                           callback):
        self._logger.debug("Creating queue factory with the following parameters:")
        self._logger.debug("rabbit_host={0}".format(str(rabbit_host)))
        self._logger.debug("rabbit_port={0}".format(str(rabbit_port)))
        self._logger.debug("rabbit_user={0}".format(str(rabbit_user)))
        self._logger.debug("rabbit_password={0}".format(str(rabbit_password)))
        self._logger.debug("exchange={0}".format(str(exchange)))
        self._logger.debug("queue_name={0}".format(str(queue_name)))
        self._logger.debug("binding_key={0}".format(str(binding_key)))
        self._logger.debug("amqp_connections={0}".format(str(amqp_connections)))
        self._logger.debug("callback={0}".format(str(callback)))
        amqp_connections = int(amqp_connections)
        queue_factory_config = {
            "queue_config": []
        }

        x = 0
        while x < amqp_connections:
            queue_config = {
                "name": "{0}_{1}".format(queue_name, str(x)),
                "type": "consumer",
                "queue": queue_name,
                "host": rabbit_host,
                "port": rabbit_port,
                "binding_key": binding_key,
                "exchange": {
                    "name": exchange,
                    "type": "x-delayed-message",
                    "arguments": {"x-delayed-type": "topic"},
                    "durable": True
                },
                "auth": {
                    "user": rabbit_user,
                    "password": rabbit_password
                }
            }
            queue_factory_config["queue_config"].append(queue_config)
            x = x + 1

        self._logger.info("Creating {0} worker threads for queue {1}..."
                          .format(str(len(queue_factory_config["queue_config"])),
                                  queue_name))
        self._logger.debug("queue_factory_config:" + str(queue_factory_config))
        queue_factory = QueueFactory()
        queue_factory.create_consumers(queue_factory_config)
        self._logger.info("Start consuming {0} worker threads for queue {1}..."
                          .format(str(len(queue_factory_config["queue_config"])),
                                  queue_name))
        queue_factory.start_consuming(self._callback_wrapper)
        self._logger.info("{0} worker threads waiting to consume messages on queue {1}..."
                          .format(str(len(queue_factory_config["queue_config"])),
                                  queue_name))
        return queue_factory

    def _new_queue_producer(self,
                            rabbit_host,
                            rabbit_port,
                            rabbit_user,
                            rabbit_password,
                            exchange,
                            binding_key):
        self._logger.debug("Creating queue producer with the following parameters:")
        self._logger.debug("rabbit_host={0}".format(str(rabbit_host)))
        self._logger.debug("rabbit_port={0}".format(str(rabbit_port)))
        self._logger.debug("rabbit_user={0}".format(str(rabbit_user)))
        self._logger.debug("rabbit_password={0}".format(str(rabbit_password)))
        self._logger.debug("exchange={0}".format(str(exchange)))
        self._logger.debug("binding_key={0}".format(str(binding_key)))

        queue_config = {
            "type": "producer",
            "host": rabbit_host,
            "port": rabbit_port,
            "binding_key": binding_key,
            "exchange": {
                "name": exchange,
            },
            "auth": {
                "user": rabbit_user,
                "password": rabbit_password
            }
        }

        return QueueProducer(queue_config)



import logging
import datetime
import json_log_formatter

class CustomisedJSONFormatter(json_log_formatter.JSONFormatter):
    def json_record(self, message: str, extra: dict, record: logging.LogRecord) -> dict:
        extra['message'] = message

        # Include builtins
        extra['severity'] = record.levelname
        extra['source'] = record.name
        extra['pathname'] = record.pathname
        extra['lineno'] = record.lineno
        extra['func'] = record.funcName

        if 'time' not in extra:
            extra['time'] = datetime.datetime.utcnow()
        if record.exc_info:
            extra['exc_info'] = self.formatException(record.exc_info)

        return extra

formatter = CustomisedJSONFormatter()
json_handler = logging.FileHandler("/dev/stdout")
json_handler.setFormatter(formatter)

# This redirects warnings to be logged, so that they are logged as json with the correct warning severity
# However, warnings that are captured by other means may need to turn this off before hand (and then back on afterwards)
logging.captureWarnings(True)
logger = logging.getLogger()
logger.addHandler(json_handler)
logger.setLevel(logging.INFO)

class LogManager:
    def __init__(self,
                 kv_store):
        # logger for this class
        self._logger = logging.getLogger("LogManager")

        self._kv_store = kv_store
        self._configure_logging(None, kv_store.get_dict(default={}))
        self._kv_store.on_reload(self._configure_logging)

    def _configure_logging(self, old_config, new_config):
        self._logger.debug("BEGIN _configure_logging")

        if not old_config or not old_config == new_config:
            self._logger.debug("Configuration changed, configuring logging...")
            logger = logging.getLogger()
            try:
                if 'logging_level' in new_config:
                    numeric_level = getattr(logging, new_config.get('logging_level'))
                    logger.setLevel(numeric_level)
                else:
                    logger.setLevel(logging.INFO)
            except:
                self._logger.warning("Invalid level, {1}, for root logging configuration."
                                     .format("logging_level", new_config.get('logging_level')))

            # Specific logger configuration
            if 'logging_config' in new_config and type(new_config.get('logging_config')) is dict:
                self._logger.debug("logging_config found.")
                logging_config = new_config.get('logging_config')

                for log_name in logging_config:
                    logger = logging.getLogger(log_name)
                    current_logging_config = logging_config.get(log_name)
                    self._logger.debug("Configuration for {0} is {1}.".format(log_name, str(current_logging_config)))

                    if type(current_logging_config) is str:
                        self._logger.debug("Configuring logger for {0}..."
                                           .format(log_name))
                        try:
                            numeric_level = getattr(logging, current_logging_config)
                            self._logger.debug("Configuring logging level to {0} for {1}..."
                                               .format(str(numeric_level), log_name))
                            logger.setLevel(numeric_level)
                        except:
                            self._logger.warning("Invalid level, {1}, for {0} logging configuration."
                                                 .format(log_name, current_logging_config))
            else:
                self._logger.debug("logging_config NOT found.")
        else:
            self._logger.debug("Configuration unchanged.")
        self._logger.debug("END _configure_logging")



import logging
from copy import deepcopy
from decimal import Decimal

from boto3.dynamodb import types

logger = logging.getLogger('utils')


def dict_replace_empty_values(in_dictionary,
                              process_none_values=False,
                              clone_dict=False,
                              remove_values=False,
                              replace_with=None,
                              replace_float_with_decimal=False):
    if type(in_dictionary) is not dict:
        raise Exception("Value provided must be a dictionary.")

    if clone_dict:
        logger.debug("Cloning dictionary")
        in_dictionary = deepcopy(in_dictionary)

    keys_to_process = []

    for key in in_dictionary.keys():
        value = in_dictionary.get(key)
        logger.debug("Processing key '{0}' of type '{1}'".format(key, type(value).__name__))
        if process_none_values and value is None:
            logger.debug("Adding key '{0}' to keys to process.".format(key))
            keys_to_process.append(key)

        if type(value) is dict:
            logger.debug("Calling dict_replace_empty_values for keys '{0}'.".format(key))
            dict_replace_empty_values(value,
                                      process_none_values=process_none_values,
                                      clone_dict=False,
                                      remove_values=remove_values,
                                      replace_with=replace_with,
                                      replace_float_with_decimal=replace_float_with_decimal)
        elif type(value) in [list, set]:
            index = 0
            if type(value) is set:
                value = list(value)
                in_dictionary[key] = value
            for item in value:
                if type(item) is dict:
                    dict_replace_empty_values(item,
                                              process_none_values=process_none_values,
                                              clone_dict=False,
                                              remove_values=remove_values,
                                              replace_with=replace_with,
                                              replace_float_with_decimal=replace_float_with_decimal)
                elif replace_float_with_decimal:
                    if type(item) is float:
                        logger.debug("Converting '{0}' to Decimal".format(key))
                        value[index] = Decimal(str(item))
                    elif type(item) is str:
                        if len(item.strip()) == 0:
                            value[index] = replace_with
                    elif type(item) in [int, Decimal, types.Binary, types.Decimal, type(None)]:
                        logger.debug("Type {0} is a valid boto3 type, not modifying value.".format(str(type(item))))
                    else:
                        logger.debug("Converting {0} to a str".format(str(type(item))))
                        value[index] = str(item)

                index = index + 1
        else:
            logger.debug("Convert non boto types? {0}".format(str(replace_float_with_decimal)))
            _process_primitives(in_dictionary, key, value, keys_to_process, replace_float_with_decimal)
            logger.debug("No special handing required for key '{0}' for its type.".format(key))

    for key_to_process in keys_to_process:
        if remove_values:
            logger.debug("Removing key '{0}'.".format(key_to_process))
            in_dictionary.pop(key_to_process)
        else:
            logger.debug("Replacing key '{0}' with '{1}'.".format(key_to_process, replace_with))
            in_dictionary[key_to_process] = replace_with

    return in_dictionary


def _process_primitives(in_dictionary, key, value, keys_to_process, replace_float_with_decimal):
    if type(value) is str:
        if len(value.strip()) == 0:
            logger.debug("Adding key '{0}' to keys to process.".format(key))
            keys_to_process.append(key)
    else:
        logger.debug("Value {0}, of type {1}, should I convert? {2}".
                     format(str(value), str(type(value)), str(replace_float_with_decimal)))
        if replace_float_with_decimal:
            logger.debug("Converting '{0}' to boto3 types".format(key))
            in_dictionary[key] = _to_boto3_type(value)
        logger.debug("Value for key '{0}' is not blank.".format(key))


def _to_boto3_type(value):
    if type(value) in [float]:
        boto3_value = Decimal(str(value))
    elif type(value) in [int, bool, types.Decimal, types.Binary, Decimal, type(None)]:
        boto3_value = value
    elif type(value) in [set, list]:
        logger.debug("Processing a {0}".format(str(type(value))))
        a_set = set()
        for i in value:
            if type[i] is str:
                if len(i.strip()) == 0:
                    logger.debug("Converting emtpy srt to None")
                    a_set.add(None)
                else:
                    logger.debug("Adding {0} to set".format(str(type(i))))
                    a_set.add(i)
            elif type(i) in [float]:
                logger.debug("Converting {0} to Decimal".format(str(type(i))))
                a_set.add(Decimal(str(i)))
            elif type(i) in [int, Decimal, types.Binary, types.Decimal]:
                logger.debug("Adding {0} to set".format(str(type(i))))
                a_set.add(i)
            else:
                logger.debug("Converting {0} to a str".format(str(type(i))))
                a_set.add(str(i))
    else:
        boto3_value = str(value)
    return boto3_value


def log_dict_types(a_dict, prefix="", types=None, use_logger=logger, print_no_type_message=False):
    was_logged = False
    for key in a_dict.keys():
        if prefix:
            fq_key = "{0}.{1}".format(prefix, key)
        else:
            fq_key = "{0}".format(key)
        value = a_dict.get(key)
        value_type = type(value).__name__
        if not types or value_type in types:
            use_logger.info("'self.{0}' is a '{1}'".format(fq_key, value_type))
            was_logged = True
        if type(value) is dict:
            log_dict_types(value, prefix=fq_key, types=types, use_logger=use_logger)
    if not was_logged:
        if prefix:
            self_prefix = "self."
        else:
            self_prefix = "self"
        if print_no_type_message:
            use_logger.info("'{0}' has no type in {1}".format("{0}{1}".format(self_prefix, prefix), str(types)))


def dig(obj, *keys):
    value = obj
    if len(keys) > 0:
        key = keys[0]
        if type(obj) is dict:
            if len(keys) == 1:
                value = obj.get(key)
            elif len(keys) > 1:
                sub_keys = tuple(keys[1:])
                value = dig(obj.get(key), *sub_keys)
        else:
            value = None
    return value

# This method returns the first machine in an INSTANCE-INFO header where the passed KV pair match or None
def authorized_machine(instance_info, machine_key, machine_value):
    try:
        return next((machine for machine in instance_info.get('machines') if machine[machine_key] == machine_value), None)
    except (TypeError, KeyError):
        return None


def read_json(file):
    logger.debug(file)
    logger.debug("======================================================")
    json_file = open(file, 'r')
    json = json_file.read()
    logger.debug(json)
    return json

def log_exception(ex):
    if ex is not None:
        logger.exception("The following exception occurred:", exc_info=ex)



import boto3
from boto3.dynamodb.conditions import Key
import logging
import simplejson as json
import os
import sys
import traceback
from clc_msa_utils.utils import authorized_machine, log_exception
from clc_msa_utils.kv_store import KVStore


class MOSDynamoClient:
    def __init__(self):
        self._kv_store = KVStore(
            kv_prefix=os.getenv('DYNAMO_ETCD_PREFIX', '/config/dynamodb')
        )
        self._dynamodb = boto3.resource('dynamodb',
                                  region_name=self._kv_store.get('db-region'),
                                  aws_access_key_id=self._kv_store.get('aws-key'),
                                  aws_secret_access_key=self._kv_store.get('aws-secret')
                                  )
        self._tables = {}
        self._tables['managed_servers_table'] = self._dynamodb.Table(self._kv_store.get('managed_servers_table'))
        self._tables['managed_applications_table'] = self._dynamodb.Table(self._kv_store.get('managed_applications_table'))
        self._tables['ad_servers_table'] = self._dynamodb.Table(self._kv_store.get('active-directory-server-table'))
        self._tables['events_table'] = self._dynamodb.Table(self._kv_store.get('events-table'))
        self._tables['networks_table'] = self._dynamodb.Table(self._kv_store.get('network-table'))
        self._tables['vpn_hubs_table'] = self._dynamodb.Table(self._kv_store.get('vpn_hubs_table'))
        self._tables['vpn_tenants_table'] = self._dynamodb.Table(self._kv_store.get('vpn_tenants_table'))
        self._tables['tickets-table'] = self._dynamodb.Table(self._kv_store.get('tickets-table'))
        self._logger = logging.getLogger("default")


    def _get_item(self, table, key):
        item = None
        try:
            response = table.get_item(Key=key)
            self._logger.debug("Response getting item by Key={}:\n{}".format(key, json.dumps(response, indent=4)))

            if "Item" in response:
                item = response['Item']

            return item
        except Exception as exception:
            self._logger.error(
                "Unexpected error occurred retrieving document from database:\n%s" % traceback.format_exc())
            self._logger.error("Exception:\n%s" % exception)
            raise exception

    def _put_item(self, table, item ):
        response = table.put_item(Item=item)
        return response['ResponseMetadata']['HTTPStatusCode']

    def _query(self, table, db_args):
        items = []
        try:
            response = table.query(**db_args)
            self._logger.debug("Response querying {}:\n{}".format(table.table_name, json.dumps(response, indent=4)))
            if response['Count'] > 0:
                items = response['Items']
        except Exception as exception:
            self._logger.error(
                "Unexpected error occurred retrieving data from {} :\n{}".format(table.table_name, traceback.format_exc()))
            self._logger.error("Exception:\n%s" % exception)
            raise exception
        return items


    def get_managed_server(self, managed_server_id, instance_info):
        self._logger.debug("Getting managed server details for server with ID = " + managed_server_id + "...")
        managed_server = self.get_managed_server_by_support_id(managed_server_id)

        if not managed_server:
            managed_server = self.get_managed_server_by_machine_name(managed_server_id, instance_info)

        self._logger.debug("Returning Managed Server:\n" + json.dumps(managed_server, indent=4))
        return managed_server

    def get_managed_server_by_support_id(self, support_id):
        self._logger.debug("Getting managed server details for server with support ID = " + support_id)
        return self._get_item(self._tables['managed_servers_table'], {'name': support_id})

    def get_managed_server_by_machine_name(self, machine_name, instance_info):
        managed_server = None
        self._logger.debug("Getting managed server details for server with machine name = " + machine_name + "...")
        db_args = {
            "KeyConditionExpression": Key('mcmInstanceId').eq(machine_name),
            "IndexName": "mcmInstanceId-index"
        }
        servers = self._query(self._tables['managed_servers_table'], db_args)
        self._logger.debug("Response getting server by machine_name:\n%s" % json.dumps(servers, indent=4))
        if len(servers) > 0:
            for server in servers:
                machine = authorized_machine(instance_info, 'support_id', server.get('name'))
                if machine:
                    managed_server = server
                    break

        return managed_server


    def get_appliances_by_network(self, vpc_id):
        db_args = {
            "KeyConditionExpression": Key('network').eq(vpc_id) & Key('isAppliance').eq('true'),
            "IndexName": "network-isAppliance-index"
        }
        return self._query(self._tables['managed_servers_table'], db_args)


    def get_managed_servers_by_gateway_name(self, gateway_name):
        db_args = {
            "KeyConditionExpression": Key('gatewayName').eq(gateway_name) & Key('status').eq('active'),
            "IndexName": "gatewayName-status-index"
        }
        return self._query(self._tables['managed_servers_table'], db_args)


    def get_managed_application(self, application_id):
        self._logger.debug("Getting managed application details for ID = " + application_id)
        return self._get_item(self.self._tables['managed_applications_table'], {'}mcmInstanceId': application_id})

    def scan_managed_servers(self, db_args):
        return self._tables['managed_servers_table'].scan(**db_args)


    def update_managed_application(self, db_args):
        return self._tables['managed_applications_table'].update_item(**db_args)


    def update_managed_server(self, db_args):
        return self._tables['managed_servers_table'].update_item(**db_args)


    def scan_ad_servers(self, db_args):
        return self._tables['ad_servers_table'].scan(**db_args)

    def put_ad_server(self, ad_server):
        self._logger.info("PUT Payload : " + str(ad_server))
        return self._put_item(self._tables['ad_servers_table'], ad_server)


    def delete_ad_server(self, gateway_name):
        try:
            response =  self._tables['ad_servers_table'].update_item(Key={'gatewayName': gateway_name},
                                                UpdateExpression="SET #status = :status",
                                                ExpressionAttributeValues={
                                                    ':status': "terminated"
                                                },
                                                ExpressionAttributeNames={
                                                    '#status': 'status'
                                                })
            return response['ResponseMetadata']['HTTPStatusCode']

        except Exception as e:
            self._logger.error("Error occurred deleting ad server with vpcId '{0}': \n{1}".format(id, str(e)))
            log_exception(sys.exc_info())
            raise e


    def get_ad_server_by_gateway_name(self, gateway_name):
        return self._get_item(self._tables['ad_servers_table'], {'gatewayName': gateway_name})


    def get_ad_servers_by_account(self, account):
        db_args = {
            "KeyConditionExpression": Key('status').eq('active') & Key('account').eq(account),
            "IndexName": "status-account-index"
        }
        return self._query(self._tables['ad_servers_table'], db_args)

    def get_ad_server_dynamo_health(self, json_data):
        table_info = {}
        table_monitor = {"name": "DynamoDB Monitor"}
        try:
            table_info["tableName"] = self._tables['ad_servers_table'].table_name
            table_info["tableStatus"] = self._tables['ad_servers_table'].table_status
            if table_info["tableStatus"] in ["CREATING", "UPDATING", "ACTIVE"]:
                table_monitor["status"] = "UP"
            else:
                table_monitor["status"] = "DOWN"
                json_data["status"] = "DOWN"
            table_monitor["tableInfo"] = table_info
        except:
            e = sys.exc_info()
            log_exception(traceback.format_exc())
            table_monitor["info"] = str(e)
            table_monitor["status"] = "DOWN"
            json_data["status"] = "DOWN"
        return table_monitor


    """
    This method is used by the network-services in two endpoints;
    PUT /network-services/public/networks which creates/updates entries in the networks table. This function is called
    and if a document is returned, it updates it, if not it creates a new one. Theoretically, only one entry should
    ever be created with a single combination of vpcId and provider_id
    
    POST /network-services/admin/networks Which actually just GETS returns the network in question
    
    Since there should only ever be a single document with the same vpcId and provider_id we throw an exception if
    there is more than one document. This may create a race condition with multiple pods of network services running
    in prod. But I would argue throw the exception will at least make us aware of an issue. But if that happens we
    will probably need to find a better way to deal with network identification.
    """
    def find_network(self, provider_network_id, provider_id):
        networks = []
        if provider_network_id and provider_id:
            self._logger.debug("Getting network details in '%s' and with provider '%s'", provider_network_id, provider_id)
            db_args = {
                "KeyConditionExpression": Key('vpcId').eq(provider_network_id) & Key('provider_id').eq(provider_id),
                "IndexName": "vpcId-provider_id-index"
            }
            networks = self._query(self._tables['networks_table'], db_args)
            if len(networks) > 1:
                raise Exception("Found more than one network for vpcId %s and provider_id %s", provider_network_id, provider_id)
        else:
            self._logger.debug("Provider network id is not given, cannot fetch details.")

        return networks[0] if networks else None

    def read_network(self, network_id):
        self._logger.debug("Getting network details in " + network_id + "...")
        network = self._get_item(self._tables['networks_table'], {'network_id': network_id})
        return network

    def get_networks_by_provider_id(self, provider_id):
        db_args = {
            "KeyConditionExpression": Key('provider_id').eq(provider_id),
            "IndexName": "provider_id-index"
        }

        return self._query(self._tables['networks_table'], db_args)



    def put_network(self, network):
        return self._put_item(self._tables['networks_table'], network)

    def check_table(self, table_key):
        table_info = {}
        table_monitor = {"name": "DynamoDB monitor"}
        try:
            table_info["tableName"] = self._tables[table_key].table_name
            table_info["tableStatus"] = self._tables[table_key].table_status
            if table_info["tableStatus"] in ["CREATING", "UPDATING", "ACTIVE"]:
                table_monitor["status"] = "UP"
            else:
                table_monitor["status"] = "DOWN"
            table_monitor["tableInfo"] = table_info
        except:
            log_exception(sys.exc_info())
            table_monitor["status"] = "DOWN"

        return table_monitor


    def get_first_managed_server_event(self, support_id):
        event = None
        db_args = {
            "KeyConditionExpression": Key('eventId').eq(support_id),
            "ProjectionExpression": "eventId,eventTime",
            "Limit": 1
        }
        events = self._query(self._tables['events_table'], db_args)
        if len(events) > 0:
            event = events[0]
            self._logger.debug("Event found: {}".format(event))

        return event

    def put_event(self, event):
        self._put_item(self._tables['events_table'], event)


    def get_vpn_hubs(self):
        db_args = {
            "KeyConditionExpression": Key('enabled').eq('true'),
            "IndexName": "enabled-index"
        }

        return self._query(self._tables['vpn_hubs_table'], db_args)


    def get_active_vpn_tenant_by_ipv6_subnet(self, subnet):
        db_args = {
            "KeyConditionExpression": Key('ipv6_prefix').eq(subnet) & Key('status').eq('active'),
            "IndexName": "ipv6_prefix-status-index",
            "Limit": 1
        }

        tenants = self._query(self._tables['vpn_tenants_table'], db_args)
        if len(tenants) > 1:
            """
            When assigning a /96 subnet to a management appliance, the subnet is chosen randomly. This query is to
            determine in the subnet chosen is already in use. There should only be a single tenant for any given /96
            subnet. We do not want to throw an exception here because returning anything but an empty array will cause
            vpn-tenant-api to chose a different subnet. We log an error though because we definitely need to know if
            there are multiple management appliances assigned the same /96 CIDR block.
            """
            self._logger.error("Multiple tenants found for {}:\n{}".format(subnet, tenants))

        return tenants

    def get_vpn_tenant(self, vpn_tenant_id):
        return self._get_item(self._tables['vpn_tenants_table'], {'vpn_tenant_id': vpn_tenant_id})

    def put_vpn_tenant(self, vpn_tenant):
        return self._put_item(self._tables['vpn_tenants_table'], vpn_tenant)


    def get_ticket(self, support_id, correlation_id):
        key = {
            'support_id': support_id,
            'correlation_id': correlation_id
        }

        return self._get_item(self._tables['tickets-table'], key)

    def update_ticket(self, db_args):
        return self._tables['tickets-table'].update_item(**db_args)


    def put_ticket(self, ticket):
        return self._put_item(self._tables['tickets-table'], ticket)

    def scan_tickets(self, db_args):
        return self.self._tables['tickets-table'].scan(**db_args)


    def query_tickets(self, db_args):
        return self._query(self._tables['tickets-table'], db_args)
